# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""Interface for classes to serialize to dictionary.

This also contains classes to ease the transition from HDF based storage to
dict based serialization.  Roughly we want to proceed as follows:
    1. Any new object directly implements :class:`.HasDict` and derives from
    :class:`.HasHDFfromDict` to be compatible to older code that still relies
    on the HDF interface.
    2. Any old object that doesn't yet directly implements :class:`.HasDict`
    can trivially derive from :class:`.HasDictfromHDF` to be compatible to
    newer code that uses the dict based serialization.
    3. Step by step we can transition old objects to directly implement
    :class:`.HasDict`.
"""

from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Any

from pyiron_base.interfaces.has_hdf import HasHDF
from pyiron_base.storage.hdfio import (
    DummyHDFio,
    _extract_module_class_name,
    _import_class,
)

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2023, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Dec 20, 2023"


def create_from_dict(obj_dict):
    """
    Create and restores an object previously written as a dictionary.

    Args:
        obj_dict (dict): must be the output of HasDict.to_dict()

    Returns:
        object: restored object
    """
    if "TYPE" not in obj_dict:
        raise ValueError(
            "invalid obj_dict! must contain type information and be the output of HasDict.to_dict!"
        )
    type_field = obj_dict["TYPE"]
    module_path, class_name = _extract_module_class_name(type_field)
    class_object = _import_class(module_path, class_name)
    version = obj_dict.get("DICT_VERSION", None)
    obj = class_object.instantiate(obj_dict, version)
    obj.from_dict(obj_dict, version)
    return obj


def _join_children_dict(children: dict[str, dict[str, Any]]) -> dict[str, Any]:
    """
    Given a nested dictionary, flatten the first level.

    >>> d = {'a': {'a1': 3}, 'b': {'b1': 4, 'b2': {'c': 42}}}
    >>> _join_children_dict(d)
    {'a/a1': 3, 'b/b1': 4, 'b/b2': {'c': 42}}

    This is intended as a utility function for nested HasDict objects, that
    to_dict their children and then want to give a flattened dict for
    writing to ProjectHDFio.write_dict_to_hdf.

    See also :func:`._split_children_dict`.
    """
    return {
        "/".join((k1, k2)): v2 for k1, v1 in children.items() for k2, v2 in v1.items()
    }


def _split_children_dict(obj_dict: dict[str, Any]) -> dict[str, Any | dict[str, Any]]:
    """
    Undoes _join_children_dict.

    Classes that use :func:`._join_children_dict` in their `_to_dict`, must
    call this function in their `_from_dict`.
    """
    subs = defaultdict(dict)
    plain = {}
    for k, v in obj_dict.items():
        if "/" not in k:
            plain[k] = v
            continue
        root, k = k.split("/", maxsplit=1)
        subs[root][k] = v
    # using update keeps type stability, i.e. we always return a plain dict
    plain.update(subs)
    return plain


def _from_dict_children(obj_dict: dict) -> dict:
    """
    Recurse through `obj_dict` and restore any objects with :class:`~.HasDict`.

    Args:
        obj_dict (dict): data previously returned from :meth:`.to_dict`
    """

    def load(inner_dict):
        # object is a not a dict, so nothing to do
        if not isinstance(inner_dict, dict):
            return inner_dict
        # if object is a dict but doesn't have type information, recurse through it to load any sub dicts that might
        if not all(k in inner_dict for k in ("NAME", "TYPE", "OBJECT", "DICT_VERSION")):
            return {k: load(v) for k, v in inner_dict.items()}
        # object has type info, so just load it
        return create_from_dict(inner_dict)

    return {k: load(v) for k, v in obj_dict.items()}


def _to_dict_children(obj_dict: dict) -> dict:
    """
    Call to_dict on any objects in the values that support it.

    Intended as a helper function for recursives object that want to to_dict
    their nested objects automatically.  It uses :func:`._join_children_dict`
    for any dictionaries returned from the children.

    The function only goes through the *first* layer of dictionary values and
    does *not* recurse through nested dictionaries.

    Args:
        obj_dict (dict): data previously returned from :meth:`._to_dict`

    Returns:
        obj_dict (dict): new dictionary with the obj_dict of the children
    """
    data_dict = {}
    child_dict = {}
    for k, v in obj_dict.items():
        if isinstance(v, HasDict):
            child_dict[k] = v.to_dict()
        elif isinstance(v, HasHDF):
            child_dict[k] = HasDictfromHDF.to_dict(v)
        else:
            data_dict[k] = v
    return data_dict | _join_children_dict(child_dict)


class HasDict(ABC):
    """
    Abstract interface to convert objects to dictionaries for storage.

    Subclasses must to implement :meth:`~._from_dict` and :meth:`~._to_dict` and may implement :meth:`.instantiate`.
    :meth:`._to_dict` is excepted to return a `dict` mapping string names to the values the object needs serialized.

    On recreating an object from scratch with :func:`.create_from_dict` first :meth:`.instantiate` is called and then
    :meth:`.from_dict` with the same `obj_dict`, i.e. it is roughly equivalent to

    >>> my_dict = dict(...)
    >>> my_object = MyType.instantiate(my_dict)
    >>> my_object.from_dict(my_dict)

    Implementations should make sure that calling `to_dict` after `from_dict` returns an equivalent dictionary even when
    the object was not obtained from :meth:`.instantiate`, such that

    >>> my_dict = dict(...)
    >>> my_object = MyType(...)
    >>> my_object.from_dict(my_dict)
    >>> my_object.to_dict() == my_dict
    True
    """

    __dict_version__ = "0.1.0"
    """A version string saved together with data returned from :meth:`._to_dict` and is passed back into
    :meth:`._from_dict`.  Implementations can use this change their representation and still read older data."""

    @classmethod
    def instantiate(cls, obj_dict: dict, version: str = None) -> "Self":
        """
        Create a blank instance of this class.

        This can be used when some values are already necessary for the objects `__init__`.

        Args:
            obj_dict (dict): data previously returned from :meth:`.to_dict`
            version (str): version tag written together with the data

        Returns:
            object: a blank instance of the object that is sufficiently initialized to call :meth:`._from_dict` on it
        """
        return cls()

    def from_dict(self, obj_dict: dict, version: str = None):
        """
        Populate the object from the serialized object.

        Args:
            obj_dict (dict): data previously returned from :meth:`.to_dict`
            version (str): version tag written together with the data
        """

        obj_dict = _split_children_dict(obj_dict)
        if version is None:
            version = obj_dict.get("DICT_VERSION", None)
        self._from_dict(obj_dict, version)

    @abstractmethod
    def _from_dict(self, obj_dict: dict, version: str = None):
        """
        Populate the object from the serialized object.

        Implementations must use :func:`._from_dict_children` if they use
        `._to_dict_children` in their implementation of :meth:`._to_dict`.

        Args:
            obj_dict (dict): data previously returned from :meth:`.to_dict`
            version (str): version tag written together with the data
        """
        pass

    def to_dict(self) -> dict:
        """
        Reduce the object to a dictionary.

        Returns:
            dict: serialized state of this object
        """
        type_dict = self._type_to_dict()
        return self._to_dict() | type_dict

    @abstractmethod
    def _to_dict(self) -> dict:
        """
        Reduce the object to a dictionary.

        Implementations may use :func:`._to_dict_children`, if they
        automatically want to call `to_dict` on any objects possible from their
        returned dictionary.

        Returns:
            dict: serialized state of this object
        """
        pass

    def _type_to_dict(self):
        # Needed for the HasDictfromHDF/HasHDFfromDict classes.  When an object
        # derives from from both them and HasHDF/HasDict it will generally need
        # HDF_VERSION and DICT_VERSION defined for the version checking inside
        # from_dict/from_hdf to work properly.  So the code below tries to
        # escalate to super in case this is the case and falls back to {} if it
        # is not
        try:
            type_dict = super()._type_to_dict()
        except AttributeError:
            type_dict = {}
        type_dict |= {
            "NAME": self.__class__.__name__,
            "TYPE": str(type(self)),
            "OBJECT": self.__class__.__name__,  # unused alias
            "DICT_VERSION": self.__dict_version__,
        }
        if hasattr(self, "__version__"):
            type_dict["VERSION"] = self.__version__
        return type_dict


class HasHDFfromDict(HasHDF, HasDict):
    """
    Implements HasHDF in terms of HasDict.

    This class is intended for "new-style" objects that are used in a context
    that only assumes that they implements HasHDF.

    Implementors may still override :meth:`.HasHDF._get_hdf_group_name`.
    """

    def _from_hdf(self, hdf, version=None):
        self.from_dict(hdf.read_dict_from_hdf(recursive=True))

    def _to_hdf(self, hdf):
        hdf.write_dict_to_hdf(self.to_dict())


class HasDictfromHDF(HasDict, HasHDF):
    """
    Implements HasDict in terms of HasHDF.

    This class is intended for "old-style" objects that should be able to be
    used as children for objects that already implement HasDict and expect
    their children to implmement it.
    """

    @classmethod
    def instantiate(cls, obj_dict: dict, version: str = None) -> "Self":
        hdf = DummyHDFio(None, "/", obj_dict)
        return cls(**cls.from_hdf_args(hdf))

    def _from_dict(self, obj_dict: dict, version: str = None):
        # DummyHDFio(project=None) looks a bit weird, but it was added there
        # only to support saving/loading jobs which already use the HasDict
        # interface
        group_name = self._get_hdf_group_name()
        if group_name is not None:
            hdf = DummyHDFio(None, "/", {group_name: obj_dict})
        else:
            hdf = DummyHDFio(None, "/", obj_dict)
        self.from_hdf(hdf)

    def _to_dict(self):
        hdf = DummyHDFio(None, "/")
        self.to_hdf(hdf)
        group_name = self._get_hdf_group_name()
        data = hdf.to_dict()
        if group_name is not None:
            return data[group_name]
        else:
            return data



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
General purpose output parser
"""

import ast
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np

__author__ = "Joerg Neugebauer"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


def extract_data_from_str_lst(
    str_lst: List[str], tag: str, num_args: int = 1
) -> List[Union[str, List[str]]]:
    """
    General purpose routine to extract any static from a log (text) file

    Args:
        str_lst (List[str]): list of strings representing the lines in the file
        tag (str): string at the beginning of the line
        num_args (int): number of arguments separated by ' ' or ',' to extract after the tag

    Returns:
        List[Union[str, List[str]]]: List of arguments extracted as strings
    """

    def multiple_delimiter_split(s: str, seps: List[str]) -> List[str]:
        res = [s]
        for sep in seps:
            s, res = res, []
            for seq in s:
                res += seq.split(sep)
        while "" in res:
            res.remove("")
        return res

    collector = []
    ind_start = len(tag.split())
    for line_in_file in str_lst:
        if line_in_file.startswith(tag):
            collector = []
            vals = multiple_delimiter_split(line_in_file, (" ", ","))
            if num_args == 1:
                collector.append(vals[ind_start])
            else:
                collector.append(vals[ind_start : num_args + ind_start])

    return collector


def extract_data_from_file(
    file_name: str, tag: str, num_args: int = 1
) -> List[Union[str, List[str]]]:
    """
    General purpose routine to extract any static from a log (text) file

    Args:
        file_name (str): file name or path to the file, can either be absolute or relative
        tag (str): string at the beginning of the line
        num_args (int): number of arguments separated by ' ' or ',' to extract after the tag

    Returns:
        List[Union[str, List[str]]]: List of arguments extracted as strings
    """
    with open(file_name) as infile:
        content = infile.readlines()
    return extract_data_from_str_lst(str_lst=content, tag=tag, num_args=num_args)


class Logstatus(object):
    """
    Generic Parser for parsing output files by searching for a specific pattern structure and extracting the data that
    follows the pattern into the status_dict dictionary.

    Args:
        iter_levels (int): Levels of iteration - default = 1
    """

    def __init__(self, h5: Optional[Any] = None, iter_levels: int = 1) -> None:
        """
        Initialize the Logstatus object.

        Args:
            h5 (Optional[Any]): HDF5 object to store the dictionary in. Defaults to None.
            iter_levels (int): Levels of iteration. Defaults to 1.
        """
        if h5 is not None:
            h5.add_group("generic")
            h5.move_up()
            self.h5 = h5
            self.h5_group_data = h5.getGroup().logStatus

        self.status_dict = {}
        self.iter_levels = iter_levels
        self.iter = iter_levels * [0]
        self.store_as_vector = []
        self.h5_open = False

    def reset_iter(self, dim: int = 0) -> None:
        """
        Reset iteration level

        Args:
            dim (int): reset value - default = 0
        """
        for i in range(dim, self.iter_levels):
            self.iter[i] = 0

    def raise_iter(self, dim: int = 0) -> None:
        """
        Increase the iteration level

        Args:
            dim (int): position - default = 0
        """
        self.iter[dim] += 1

    def append(
        self, title: str, data_to_append: Union[list, dict], vec: bool = False
    ) -> None:
        """
        Append data to the LogStatus object status_dict dictionary

        Args:
            title (str): Title of the data to append
            data_to_append (Union[list, dict]): the data can be of various types
            vec (bool): [True/False] if the data is a single vector instead of a matrix or a tensor
        """
        if title in self.status_dict.keys():
            if vec:
                raise ValueError(
                    "For appending matrix rather than vector option needed!"
                )
            self.status_dict[title].append([list(self.iter), data_to_append])
        else:
            self.status_dict[title] = [[list(self.iter), data_to_append]]

    def to_hdf(self, hdf: "ProjectHDFio") -> None:
        """
        Store the LogStatus object status_dict dictionary in an HDF5 file

        Args:
            hdf (ProjectHDFio): HDF5 object to store the dictionary in.
        """
        for key, value in self.status_dict.items():
            if key in self.store_as_vector:
                if len(value) > 1:
                    raise ValueError(
                        "Multi-dimensional array cannot be saved as vector"
                    )
                hdf[key] = np.array(value[0][1])
            else:
                hdf[key] = np.array([val for _, val in value])

    def combine_xyz(
        self,
        x_key: str,
        y_key: str,
        z_key: str,
        combined_key: str,
        as_vector: bool = False,
    ) -> None:
        """
        Combine three lists representing the x,y,z coordinates, by accessing them from the status_dict dictionary,
        combining them, store them under the combined_key and remove the other three keys.

        Args:
            x_key (str): key of the x coordinates
            y_key (str): key of the y coordinates
            z_key (str): key of the z coordinates
            combined_key (str): name of the combined coordinates
            as_vector (bool): [True/False] if the combined coordinates should be stored as a single vector instead of a matrix. Defaults to False.
        """
        if (
            x_key in self.status_dict
            and y_key in self.status_dict
            and z_key in self.status_dict
        ):
            combined_lst = []
            if as_vector:
                time_x, val_x = self.status_dict[x_key][0]
                time_y, val_y = self.status_dict[y_key][0]
                time_z, val_z = self.status_dict[z_key][0]
                for val_t_x, val_t_y, val_t_z in zip(val_x, val_y, val_z):
                    combined_lst.append([time_x, [val_t_x, val_t_y, val_t_z]])
            else:
                for var_x, var_y, var_z in zip(
                    self.status_dict[x_key],
                    self.status_dict[y_key],
                    self.status_dict[z_key],
                ):
                    time_x, val_x = var_x
                    time_y, val_y = var_y
                    time_z, val_z = var_z
                    combined_lst.append(
                        [
                            time_x,
                            [
                                [val_t_x, val_t_y, val_t_z]
                                for val_t_x, val_t_y, val_t_z in zip(
                                    val_x, val_y, val_z
                                )
                            ],
                        ]
                    )
            del self.status_dict[x_key]
            del self.status_dict[y_key]
            del self.status_dict[z_key]
            self.status_dict[combined_key] = combined_lst

    def combine_mat(
        self,
        x_key: str,
        xy_key: str,
        xz_key: str,
        y_key: str,
        yz_key: str,
        z_key: str,
        combined_key: str,
    ) -> None:
        """
        Combine three lists representing the x,y,z coordinates, by accessing them from the status_dict dictionary,
        combining them, store them under the combined_key and remove the other three keys.

        Args:
            x_key (str): key of the x coordinates
            xy_key (str): key of the xy coordinates
            xz_key (str): key of the xz coordinates
            y_key (str): key of the y coordinates
            yz_key (str): key of the yz coordinates
            z_key (str): key of the z coordinates
            combined_key (str): name of the combined coordinates
        """
        if (
            x_key in self.status_dict
            and y_key in self.status_dict
            and z_key in self.status_dict
        ):
            combined_lst = []
            for var_xx, var_xy, var_xz, var_yy, var_yz, var_zz in zip(
                self.status_dict[x_key],
                self.status_dict[xy_key],
                self.status_dict[xz_key],
                self.status_dict[y_key],
                self.status_dict[yz_key],
                self.status_dict[z_key],
            ):
                time_xx, val_xx = var_xx
                time_xy, val_xy = var_xy
                time_xz, val_xz = var_xz
                time_yy, val_yy = var_yy
                time_yz, val_yz = var_yz
                time_zz, val_zz = var_zz
                combined_lst.append(
                    [
                        time_xx,
                        [
                            [
                                [var_t_xx, var_t_xy, var_t_xz],
                                [var_t_yx, var_t_yy, var_t_yz],
                                [var_t_zx, var_t_zy, var_t_zz],
                            ]
                            for var_t_xx, var_t_xy, var_t_xz, var_t_yx, var_t_yy, var_t_yz, var_t_zx, var_t_zy, var_t_zz in zip(
                                val_xx,
                                val_xy,
                                val_xz,
                                val_xy,
                                val_yy,
                                val_yz,
                                val_xz,
                                val_yz,
                                val_zz,
                            )
                        ],
                    ]
                )
            del self.status_dict[x_key]
            del self.status_dict[xy_key]
            del self.status_dict[xz_key]
            del self.status_dict[y_key]
            del self.status_dict[yz_key]
            del self.status_dict[z_key]
            self.status_dict[combined_key] = combined_lst

    def convert_unit(self, key: str, factor: float) -> None:
        """
        Convert the values of a specific key in the status_dict dictionary by multiplying them with a factor.

        Args:
            key (str): The key of the values to be converted.
            factor (float): The factor to multiply the values with.
        """
        if key in self.status_dict:
            return_lst = []
            for step in self.status_dict[key]:
                time, values = step
                return_lst.append([time, (np.array(values) * factor).tolist()])
            self.status_dict[key] = return_lst

    @staticmethod
    def extract_item(l_item: str) -> Tuple[str, Optional[List[str]]]:
        """
        Method to extract information from a single line - currently very specific for the Lammps output

        Args:
            l_item (str): line to extract information from

        Returns:
            Tuple[str, Optional[List[str]]]: the tag_string as string and the arguments as list
        """
        item_list = l_item.split()
        first_item = item_list[1]
        if first_item == "NUMBER":
            num_elements = 3
        elif first_item == "BOX":
            num_elements = 2
        else:
            num_elements = 1
        tag = item_list[1 : num_elements + 1]
        tag_string = " ".join(el for el in tag)
        if len(item_list) == num_elements + 1:
            args = None
        else:
            args = item_list[num_elements + 1 : :]
        return tag_string, args

    def extract_from_list(
        self,
        list_of_lines: List[str],
        tag_dict: Dict[str, Any],
        h5_dict: Optional[Dict[str, str]] = None,
        key_dict: Optional[Dict[str, str]] = None,
    ) -> None:
        """
        Main function of the LogStatus class to extract data from an output file by searching for the tag dictionary

        Args:
            list_of_lines (List[str]): List of lines from the output file
            tag_dict (Dict[str, Any]): Dictionary with tags/patterns as key and an additional dictionary to describe the data structure.
            h5_dict (Optional[Dict[str, str]]): Translation dictionary of output tags as keys to the tags used on the HDF5 file as values. Defaults to None.
            key_dict (Optional[Dict[str, str]]): Translation dictionary of python internal tags as keys to the output tags as values. Defaults to None.
        """
        val_item = {}
        tag_vals = {}

        tag = LogTag(tag_dict, h5_dict, key_dict)

        iterate_over_lines = iter(list_of_lines)
        for line_read in iterate_over_lines:
            while True:
                if tag.is_item(line_read):  # items):
                    tag_name = tag.tag_name
                    if tag.rows() == 0:  # read single line_read
                        tag.set_item(tag_vals, self)
                        try:
                            line_read = next(iterate_over_lines)
                        except StopIteration:
                            break
                    else:
                        for _ in range(tag.line_skip()):
                            line_read = next(iterate_over_lines)
                        if isinstance(tag.rows(), str):
                            i_line = 0
                            while True:
                                try:
                                    line_read = next(iterate_over_lines)
                                except StopIteration:
                                    break
                                if line_read.find(tag.rows().strip()) > -1:
                                    break
                                if "WARNING:" in line_read:
                                    break
                                val_line = [
                                    [
                                        ast.literal_eval(line)
                                        for line in line_read.split()
                                    ]
                                ]
                                if i_line == 0:
                                    val_array = np.array(val_line)
                                else:
                                    val_array = np.append(
                                        arr=val_array, values=val_line, axis=0
                                    )
                                i_line += 1

                        else:
                            for i_line in range(tag.rows()):
                                try:
                                    line_read = next(iterate_over_lines)
                                except StopIteration:
                                    break
                                val_line = [
                                    [
                                        ast.literal_eval(line)
                                        for line in line_read.split()
                                    ]
                                ]
                                if i_line == 0:
                                    val_array = np.array(val_line)
                                else:
                                    val_array = np.append(
                                        arr=val_array, values=val_line, axis=0
                                    )

                        if tag.is_func():
                            val_array = tag.apply_func(val_array)

                        val_item[tag_name] = val_array
                        if np.shape(val_array) == (1, 1):
                            self.append(tag.h5(), val_array[0, 0])
                        elif tag.test_split():
                            tag_list = None
                            if tag.split_tag:
                                tag_list = tag_name.split()
                            elif tag.split_arg:
                                if "header" not in tag_dict[tag_name].keys():
                                    tag_list = tag.val_list
                                else:
                                    tag_list = tag_dict[tag_name]["header"]
                            for i, t in enumerate(tag_list):
                                if "header" not in tag_dict[tag_name].keys():
                                    self.append(
                                        tag.translate(t), np.copy(val_array[:, i])
                                    )
                                else:
                                    self.append(t, np.copy(val_array[:, i]))
                        else:
                            self.append(tag.h5(), np.copy(val_array))
                else:
                    try:
                        line_read = next(iterate_over_lines)
                    except StopIteration:
                        break

    def extract_file(
        self,
        file_name: str,
        tag_dict: Dict[str, Any],
        h5_dict: Optional[Dict[str, str]] = None,
        key_dict: Optional[Dict[str, str]] = None,
    ) -> None:
        """
        Main function of the LogStatus class to extract data from an output file by searching for the tag dictionary

        Args:
            file_name (str): absolute path to the output file
            tag_dict (dict): Dictionary with tags/patterns as key and an additional dictionary to describe the data
                             structure. The data structure dictionary can contain the following keys:
                             - "arg": position of the argument - or dimension (":", ":,:")
                             - "type": Python data type
                             - "h5": HDF5 key to store the information
                             - "rows": number of rows from the line where the tag was found
                             - "splitTag": split the tag - [True/False]
                             - "splitArg": split the argument - [True/False]
                             - "lineSkip": skip a line
                             - "func": function to convert the data
            h5_dict (dict): Translation dictionary of output tags as keys to the tags used on the HDF5 file as values.
            key_dict (dict): Translation dictionary of python internal tags as keys to the output tags as values.
        """
        with open(file_name, "r") as f:
            content = f.readlines()
        self.extract_from_list(
            list_of_lines=content, tag_dict=tag_dict, h5_dict=h5_dict, key_dict=key_dict
        )


class LogTag(object):
    """
    LogTag object to parse for a specific pattern in the output file

    Args:
        tag_dict (Dict[str, Any]): Dictionary with tags/patterns as key and an additional dictionary to describe the data
                                   structure. The data structure dictionary can contain the following keys:
                                    - "arg": position of the argument - or dimension (":", ":,:")
                                    - "type": Python data type
                                    - "h5": HDF5 key to store the information
                                    - "rows": number of rows from the line where the tag was found
                                    - "splitTag": split the tag - [True/False]
                                    - "splitArg": split the argument - [True/False]
                                    - "lineSkip": skip a line
                                    - "func": function to convert the data
        h5_dict (Optional[Dict[str, str]]): Translation dictionary of output tags as keys to the tags used on the HDF5 file as values. Defaults to None.
        key_dict (Optional[Dict[str, str]]): Translation dictionary of python internal tags as keys to the output tags as values. Defaults to None.
    """

    def __init__(
        self,
        tag_dict: Dict[str, Any],
        h5_dict: Optional[Dict[str, str]] = None,
        key_dict: Optional[Dict[str, str]] = None,
    ) -> None:
        self._tag_dict = None
        self._tag_first_word = None
        self._current = None
        self._dyn_tags = None
        self._key_dict = None
        self._h5_dict = None
        self._tag_name = None
        self.tag_dict = tag_dict
        self.key_dict = key_dict
        self.h5_dict = h5_dict

    @property
    def current(self) -> Dict[str, Any]:
        """
        Get the current tag

        Returns:
            dict: current tag
        """
        return self._current

    @current.setter
    def current(self, tag_name: str) -> None:
        """
        Set the current tag

        Args:
            tag_name (str): current tag
        """
        if tag_name not in self.tag_dict.keys():
            raise ValueError("Unknown tag_name: " + tag_name)
        self._tag_name = tag_name
        self._current = self.tag_dict[tag_name]

    @property
    def tag_name(self) -> str:
        """
        Get tag name

        Returns:
            str: tag name
        """
        return self._tag_name

    @property
    def tag_dict(self) -> dict:
        """
        Get tag dictionary with tags/patterns as key and an additional dictionary to describe the data
        structure. The data structure dictionary can contain the following keys:
        - "arg": position of the argument - or dimension (":", ":,:")
        - "type": Python data type
        - "h5": HDF5 key to store the information
        - "rows": number of rows from the line where the tag was found
        - "splitTag": split the tag - [True/False]
        - "splitArg": split the argument - [True/False]
        - "lineSkip": skip a line
        - "func": function to convert the data

        Returns:
            dict: tag dictionary
        """
        return self._tag_dict

    @tag_dict.setter
    def tag_dict(self, tag_dict: dict) -> None:
        """
        Set tag dictionary with tags/patterns as key and an additional dictionary to describe the data
        structure. The data structure dictionary can contain the following keys:
        - "arg": position of the argument - or dimension (":", ":,:")
        - "type": Python data type
        - "h5": HDF5 key to store the information
        - "rows": number of rows from the line where the tag was found
        - "splitTag": split the tag - [True/False]
        - "splitArg": split the argument - [True/False]
        - "lineSkip": skip a line
        - "func": function to convert the data

        Args:
            tag_dict (dict): tag dictionary
        """
        self._tag_dict = tag_dict
        self._tag_first_word = tuple(self.tag_dict.keys())
        self.dyn_tags = tag_dict

    @property
    def tag_first_word(self) -> str:
        """
        Get first word of the tag

        Returns:
            str: first word
        """
        return self._tag_first_word

    @property
    def dyn_tags(self) -> dict:
        """
        Get dynamic tags

        Returns:
            dict: dynamic tags
        """
        return self._dyn_tags

    @dyn_tags.setter
    def dyn_tags(self, tag_dict: dict) -> None:
        """
        Set dynamic tags

        Args:
            tag_dict (dict): tag dictionary
        """
        dyn_tags = {}
        for w in tag_dict.keys():
            items = w.split()
            if items[0][:1] == "$":
                dyn_tags[w[1:]] = w
        self._dyn_tags = dyn_tags

    @property
    def key_dict(self) -> dict:
        """
        Get translation dictionary of python internal tags as keys to the output tags as values.

        Returns:
            dict: key dictionary
        """
        return self._key_dict

    @key_dict.setter
    def key_dict(self, key_dict: dict) -> None:
        """
        Set translation dictionary of python internal tags as keys to the output tags as values.

        Args:
            key_dict (dict): key dictionary
        """
        self._key_dict = key_dict

    @property
    def h5_dict(self) -> dict:
        """
        Get translation dictionary of output tags as keys to the tags used on the HDF5 file as values.

        Returns:
            dict: h5 dictionary
        """
        return self._h5_dict

    @h5_dict.setter
    def h5_dict(self, h5_dict: dict) -> None:
        """
        Set translation dictionary of output tags as keys to the tags used on the HDF5 file as values.

        Args:
            h5_dict (dict): h5 dictionary
        """
        self._h5_dict = h5_dict

    def is_item(self, item_line: str, start: int = 0) -> bool:
        """
        Check if the current line - item_line - matches one of the provided tags, if that is the case set the tag to be
        the current tag and update the val_list with the corresponding values.

        Args:
            item_line (str): Line of the output file
            start (int): Character to start with when parsing the item_line - default=0

        Returns:
            bool: [True/False]
        """
        line = item_line.strip()
        if not line.startswith(
            self.tag_first_word, start
        ):  # start -> line must start with tag
            return False
        tag = None
        for tag in self.tag_first_word:
            if start == line.find(tag, start):
                break

        items = [ls.strip() for ls in line[len(tag) :].split()]
        self.current = tag
        self.val_list = items
        return True

    def get_item(self, item: str, default: Any) -> Union[list, dict, int, float]:
        """
        If item is part of the current dictionary keys the corresponding value is returned otherwise the default is
        returned.

        Args:
            item (str): dictionary key
            default (list, dict, int, float): Default value

        Returns:
            list, dict, int, float: The values connected to the key item in the current dictionary and if item is not a
                                    key in the current dictionary return the default value.
        """
        if self.current is None:
            raise ValueError("current tag not defined!")
        if item in self.current.keys():
            return self.current[item]
        else:
            return default

    def h5(self) -> str:
        """
        Translate current tag to HDF5 tag using the tag dictionary

        Returns:
            str: hdf5 key name
        """
        return self.get_item(item="h5", default=self.tag_name)

    def translate(self, item: str) -> str:
        """
        Translate current tag to HDF5 tag using the h5_dict dictionary

        Args:
            item (str): Python tag

        Returns:
            str: HDF5 tag
        """
        if self.h5_dict is None:
            raise ValueError("h5_dict is None!" + item)
        if item in self.h5_dict.keys():
            return self.h5_dict[item]
        else:
            raise ValueError("tag not in h5_dict: " + item)

    def arg(self) -> str:
        """
        Get tag argument

        Returns:
            str: tag arguments
        """
        l_arg = self.get_item(item="arg", default=0)
        if isinstance(l_arg, str):
            return l_arg
        else:
            return str(l_arg)

    def line_skip(self) -> bool:
        """
        Check how many lines should be skipped.

        Returns:
            bool: [True/ False]
        """
        return bool(self.get_item(item="lineSkip", default=0))

    def rows(self) -> Union[int, str]:
        """
        Number of rows to parse

        Returns:
            int, str: number of rows
        """
        rows = self.get_item(item="rows", default=0)
        try:
            return int(rows)
        except ValueError:
            return rows

    def test_split(self) -> bool:
        """
        Check if the argument or the tag should be split - if "splitArg" or "splitTag" is included in the tag_dict
        dictionary.

        Returns:
            bool: [True/ False]
        """
        self.split_arg = self.get_item(item="splitArg", default=False)
        self.split_tag = self.get_item(item="splitTag", default=False)
        return self.split_arg or self.split_tag

    def is_func(self) -> bool:
        """
        Check if a function is defined to convert the data - if "func" is included in the tag_dict dictionary

        Returns:
            bool: [True/ False]
        """
        my_func = self.get_item(item="func", default=None)
        return my_func is not None

    def apply_func(
        self, val: Union[list, dict, int, float]
    ) -> Union[list, dict, int, float]:
        """
        Apply the function on a given value

        Args:
            val (dict, list, float, int): value to apply the function on

        Returns:
            dict, list, float, int: result of applying the function
        """
        my_func = self.get_item(item="func", default=None)
        if my_func is not None:
            return my_func(val)

    def set_item(
        self, tag_vals: dict, log_file: Logstatus
    ) -> Tuple[str, dict, int, bool]:
        """
        Set LogTag item

        Args:
            tag_vals (dict): tag value dictionary
            log_file (Logstatus): Logstatus object

        Returns:
            list: tag name, tag values, rows, line skip [True/False]
        """
        tag_name = self.tag_name
        if self.rows() == 0:
            if not len(self.arg()) == 1:
                val = []
                for i_item in ast.literal_eval(self.arg()):
                    val.append(ast.literal_eval("self.val_list[" + i_item + "]"))
            else:  # input is an array
                val = eval("self.val_list[" + self.arg() + "]")
            if isinstance(val, str):
                val = ast.literal_eval(val)
            tag_vals[tag_name] = val
            if len(self.arg()) == 1:
                log_file.append(self.h5(), data_to_append=val)
            else:
                for i_num, i_val in enumerate(val):
                    log_file.append(self.h5()[i_num], data_to_append=i_val)
            if tag_name in self.dyn_tags.keys():
                self.resolve_dynamic_variable(val)
        return tag_name, tag_vals, self.rows(), self.line_skip()

    def resolve_dynamic_variable(self, val: list) -> None:
        """
        Resolve dynamic variable using the key_dict dictionary

        Args:
            val: values to resolve
        """
        d_name = self.dyn_tags[self.tag_name]
        if self.key_dict is not None:
            val = [self.key_dict[v] for v in val if v in self.key_dict.keys()]
        resolved_name = " ".join(val)

        v = self.tag_dict[d_name]
        self.tag_dict[resolved_name] = v
        del self.tag_dict[d_name]
        self.dyn_tags = self.tag_dict
        self._tag_first_word = tuple(self.tag_dict.keys())



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
The job wrapper is called from the run_job.py script, it restores the job from hdf5 and executes it.
"""

import logging
import os
from typing import Optional

from pyiron_base.database.filetable import (
    get_hamilton_from_file,
    get_hamilton_version_from_file,
    get_job_status_from_file,
)
from pyiron_base.project.generic import Project
from pyiron_base.state import state
from pyiron_base.state.signal import catch_signals

__author__ = "Joerg Neugebauer"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


class JobWrapper(object):
    """
    The job wrapper is called from the run_job.py script, it restores the job from hdf5 and executes it.

    Args:
        working_directory (str): working directory of the job
        job_id (int/ None): job ID
        hdf5_file (str): path to the HDF5 file of the job
        h5_path (str): path inside the HDF5 file to load the job
        submit_on_remote (bool): submit to queuing system on remote host
        debug (bool): enable debug mode [True/False] (optional)
    """

    def __init__(
        self,
        working_directory: str,
        job_id: Optional[int] = None,
        hdf5_file: Optional["pyiron_base.storage.hdfio.ProjectHDFio"] = None,
        h5_path: Optional[str] = None,
        submit_on_remote: bool = False,
        debug: bool = False,
        connection_string: Optional[str] = None,
        collect: bool = False,
    ):
        self.working_directory = working_directory
        self._remote_flag = submit_on_remote
        self._collect = collect
        if connection_string is not None:
            state.database.open_local_sqlite_connection(
                connection_string=connection_string
            )
        pr = Project(path=os.path.join(working_directory, "..", ".."))
        if job_id is not None:
            self.job = pr.load(int(job_id))
        else:
            projectpath = state.database.top_path(hdf5_file)
            if projectpath is None:
                project = os.path.dirname(hdf5_file)
            else:
                project = os.path.relpath(os.path.dirname(hdf5_file), projectpath)
            job_name = h5_path[1:]
            self.job = pr.load_from_jobpath(
                job_id=None,
                db_entry={
                    "job": job_name,
                    "subjob": h5_path,
                    "projectpath": projectpath,
                    "project": project + "/",
                    "status": get_job_status_from_file(
                        hdf5_file=hdf5_file, job_name=job_name
                    ),
                    "hamilton": get_hamilton_from_file(
                        hdf5_file=hdf5_file, job_name=job_name
                    ),
                    "hamversion": get_hamilton_version_from_file(
                        hdf5_file=hdf5_file, job_name=job_name
                    ),
                },
                convert_to_object=True,
            )

        # setup logger
        self._logger = self.setup_logger(debug=debug)

    @staticmethod
    def setup_logger(debug: bool = False) -> logging.Logger:
        """
        Setup the error logger

        Args:
            debug (bool): the level of logging, enable debug mode [True/False] (optional)

        Returns:
            logger: logger object instance
        """
        logger = logging.getLogger("pyiron_log")
        logger.setLevel(logging.INFO)
        if debug:
            logger.setLevel(logging.DEBUG)
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        ch.setFormatter(formatter)
        logger.addHandler(ch)
        return logger

    def run(self) -> None:
        """
        The job wrapper run command, sets the job status to 'running' and executes run_if_modal().
        """
        if self._remote_flag and self.job.server.queue is not None:
            self.job.run_if_scheduler()
            self.job.status.submitted = True
        elif self._collect:
            self.job.status.collect = True
            self.job.run()
        else:
            with catch_signals(self.job.signal_intercept):
                self.job.run_static()


def job_wrapper_function(
    working_directory: str,
    job_id: Optional[int] = None,
    file_path: Optional[str] = None,
    submit_on_remote: bool = False,
    debug: bool = False,
    collect: bool = False,
):
    """
    Job Wrapper function - creates a JobWrapper object and calls run() on that object

    Args:
        working_directory (str): directory where the HDF5 file of the job is located
        job_id (int/ None): job id
        file_path (str): path to the HDF5 file
        debug (bool): enable debug mode
        submit_on_remote (bool): submit to queuing system on remote host
        collect (bool): collect output of calculation
    """

    # always close the database connection in calculations on the cluster to avoid high number of concurrent
    # connections.
    state.database.close_connection()
    state.database.connection_timeout = 0
    state.database.open_connection()

    if job_id is not None:
        job = JobWrapper(
            working_directory=working_directory,
            job_id=job_id,
            submit_on_remote=submit_on_remote,
            debug=debug,
            collect=collect,
        )
    elif file_path is not None:
        hdf5_file = (
            ".".join(file_path.split(".")[:-1])
            + "."
            + file_path.split(".")[-1].split("/")[0]
        )
        h5_path = "/".join(file_path.split(".")[-1].split("/")[1:])
        job = JobWrapper(
            working_directory,
            job_id=None,
            hdf5_file=hdf5_file,
            h5_path="/" + h5_path,
            submit_on_remote=submit_on_remote,
            debug=debug,
            collect=collect,
        )
    else:
        raise ValueError("Either job_id or file_path have to be not None.")
    job.run()



import os
import subprocess
import warnings

from conda.core.envs_manager import list_all_known_prefixes


class CondaEnvironment:
    def __init__(self, env_path: str):
        """
        Initialize a CondaEnvironment object.

        Args:
            env_path (str): The path to the conda environment.
        """
        self._env_path = env_path

    def __dir__(self):
        return list(self._list_all_known_prefixes_dict().keys()) + ["create"]

    @staticmethod
    def _list_all_known_prefixes_dict() -> dict:
        """
        Return a dictionary of all known conda environment prefixes.

        Returns:
            dict: A dictionary of conda environment prefixes.
        """
        return {os.path.basename(path): path for path in list_all_known_prefixes()}

    def __getattr__(self, item: str) -> str:
        item_dict = {os.path.basename(path): path for path in list_all_known_prefixes()}
        if item in item_dict.keys():
            return item_dict[item]
        else:
            raise AttributeError(
                f"Unknown conda environment {item}. Use one of {self._list_all_known_prefixes_dict()} or create a new one."
            )

    def create(
        self,
        env_name: str,
        env_file: str,
        use_mamba: bool = False,
        global_installation: bool = True,
    ) -> None:
        """
        Create a new conda environment.

        Args:
            env_name (str): The name of the new environment.
            env_file (str): The path to the environment file.
            use_mamba (bool, optional): Whether to use mamba instead of conda. Defaults to False.
            global_installation (bool, optional): Whether to install the environment globally. Defaults to True.

        Raises:
            subprocess.CalledProcessError: If the environment creation command fails.
        """
        exe = "mamba" if use_mamba else "conda"
        env_lst = list_all_known_prefixes()
        env_path = os.path.join(os.path.abspath(self._env_path), env_name)
        if env_name not in env_lst and env_path not in env_lst:
            command_lst = [
                exe,
                "env",
                "create",
            ]
            if not global_installation:
                os.makedirs(self._env_path, exist_ok=True)
                command_lst += [
                    "--prefix",
                    env_path,
                ]
            command_lst += [
                "-f",
                env_file,
                "-y",
            ]
            subprocess.check_output(
                command_lst,
                universal_newlines=True,
            )
        else:
            warnings.warn(
                "The conda environment "
                + env_name
                + " already exists in "
                + str(env_path)
                + "."
            )



import os
from itertools import islice
from typing import Generator, List, Optional, Union

from pyiron_base.jobs.job.util import (
    _working_directory_list_files,
    _working_directory_read_file,
)


class File:
    __slots__ = ("_path",)

    def __init__(self, path: str):
        self._path = path

    def __str__(self) -> str:
        return self._path

    def _read(self, tail: Optional[int] = None) -> List[str]:
        return _working_directory_read_file(
            working_directory=os.path.dirname(str(self)),
            file_name=os.path.basename(str(self)),
            tail=tail,
        )

    def __iter__(self) -> Generator:
        return iter(self._read())

    def abspath(self) -> str:
        """
        Absolute path to file object

        Returns:
            str: absolute path
        """
        return os.path.abspath(self._path)

    def list(self, lines: Optional[int] = None) -> List[str]:
        """
        Return file content as list of lines.

        Args:
            lines (int): only return the first `lines` lines

        Return:
            list of str: file content
        """
        return list(islice(iter(self), lines))

    def tail(self, lines: int = 100) -> None:
        """
        Print the last `lines` to stdout.

        Args:
            lines (int): number of output lines
        """
        print(*self._read(tail=lines), sep="")

    def __eq__(self, other: Union[str, "File"]) -> bool:
        return self.__str__().__eq__(other)


class FileBrowser:
    """
    Allows to browse the files in a job directory.

    By default this object prints itself as a listing of the job directory and
    the files inside.

    >>> job.files
    /path/to/my/job:
    \tpyiron.log
    \terror.out

    Access to the names of files is provided with :meth:`.list`

    >>> job.files.list()
    ['pyiron.log', 'error.out', 'INCAR']

    Access to the contents of files is provided by indexing into this object,
    which returns a list of lines in the file

    >>> job.files['error.out']
    ["Oh no\n", "Something went wrong!\n"]

    The :meth:`.tail` method prints the last lines of a file to stdout

    >>> job.files.tail('error.out', lines=1)
    Something went wrong!

    For files that have valid python variable names can also be accessed by
    attribute notation

    >>> job.files.INCAR # doctest: +SKIP
    File('INCAR')
    """

    __slots__ = ("_working_directory",)

    def __init__(self, working_directory: str):
        self._working_directory = os.path.abspath(os.path.expanduser(working_directory))

    def _get_file_dict(self) -> dict:
        return {
            f.replace(".", "_"): f
            for f in _working_directory_list_files(
                working_directory=self._working_directory
            )
        }

    def __dir__(self) -> List[str]:
        return list(self._get_file_dict().keys()) + super().__dir__()

    def list(self) -> List[str]:
        """
        List all files in the working directory of the job.
        """
        return _working_directory_list_files(working_directory=self._working_directory)

    def _ipython_display_(self) -> None:
        path = self._job.working_directory + ":"
        files = [
            "\t" + str(f)
            for f in _working_directory_list_files(
                working_directory=self._working_directory
            )
        ]
        print(os.linesep.join([path, *files]))

    def tail(self, file: str, lines: int = 100) -> None:
        """
        Print the last lines of a file.

        Args:
            file (str): filename
            lines (int): number of lines to print

        Raises:
            FileNotFoundError: if the given file does not exist
        """
        return self[file].tail(lines=lines)

    def __getitem__(self, item: str) -> Union[File, "FileBrowser"]:
        sub = os.path.join(self._working_directory, item)
        if os.path.isdir(sub):
            return FileBrowser(sub)
        if item in _working_directory_list_files(
            working_directory=self._working_directory,
            include_archive=False,
        ):
            return File(os.path.join(self._working_directory, item))
        elif item in _working_directory_list_files(
            working_directory=self._working_directory,
            include_archive=True,
        ):
            return File(os.path.join(self._working_directory, item))
        else:
            raise FileNotFoundError(item)

    def __getattr__(self, item: str) -> Union[File, "FileBrowser"]:
        if item.startswith("__") and item.endswith("__"):
            raise AttributeError(item)
        else:
            try:
                return self[self._get_file_dict()[item]]
            except KeyError:
                raise FileNotFoundError(item) from None



import logging
import os
from typing import List


def warn_dynamic_job_classes(
    resource_folder_lst: List[str], logger: logging.Logger
) -> None:
    """
    Warns about deprecated 'dynamic' and 'templates' folders in the resource directory.

    Args:
        resource_folder_lst (List[str]): List of resource folder paths.
        logger (logging.Logger): Logger object for logging warning messages.

    Returns:
        None
    """
    for path in resource_folder_lst:
        if os.path.exists(path):
            sub_folders = os.listdir(path)
            for folder in ["dynamic", "templates"]:
                if folder in sub_folders:
                    logger.warning(
                        "pyiron found a '"
                        + folder
                        + "' folder in the "
                        + path
                        + " resource directory. These are no longer supported in pyiron_base >=0.7.0. "
                        + "They are replaced by Project.create_job_class() and Project.wrap_python_function()."
                    )



DEFAULT_STR_LENGTH = 255
PROJECT_PATH_STR_LENGTH = DEFAULT_STR_LENGTH
PROJECT_STR_LENGTH = DEFAULT_STR_LENGTH
JOB_STR_LENGTH = DEFAULT_STR_LENGTH
SUBJOB_STR_LENGTH = DEFAULT_STR_LENGTH
CHEMICALFORMULA_STR_LENGTH = 50
STATUS_STR_LENGTH = 20
HAMILTON_STR_LENGTH = 20
HAMVERSION_STR_LENGTH = 50
USERNAME_STR_LENGTH = 20
COMPUTER_STR_LENGTH = 100



import hashlib
import inspect
import re
from typing import Dict, List, Optional, Tuple

import cloudpickle

from pyiron_base.jobs.job.extension.server.generic import Server


def draw(node_dict: Dict[str, object], edge_lst: List[List[str]]) -> None:
    """
    Draw graph of nodes and edges

    Args:
        node_dict (Dict[str, object]): Dictionary of nodes
        edge_lst (List[List[str]]): List of edges
    """
    import networkx as nx
    from IPython.display import SVG, display

    graph = nx.DiGraph()
    for k, v in node_dict.items():
        graph.add_node(k, label=str(k).rsplit("_", 1)[0] + "=" + str(v))
    for edge in edge_lst:
        graph.add_edge(edge[1], edge[0])
    svg = nx.nx_agraph.to_agraph(graph).draw(prog="dot", format="svg")
    display(SVG(svg))


def evaluate_function(funct: callable, input_dict: dict) -> object:
    """
    Evaluate python function by using the input dictionary as *args and **kwargs

    Args:
        funct (callable): the function to evaluate
        input_dict (dict): input dictionary

    Returns:
        object: output of the evaluated function
    """
    input_dict = recursive_dict_resolve(input_dict=input_dict)
    if "kwargs" in input_dict.keys():
        input_dict.update(input_dict.pop("kwargs"))
    if "args" in input_dict.keys():
        args = input_dict.pop("args")
        return funct(*args, **input_dict)
    else:
        return funct(**input_dict)


def get_function_parameter_dict(funct: callable) -> dict:
    """
    Get dictionary of parameters for a function

    Args:
        funct (callable): the function to get the parameters for

    Returns:
        dict: parameters for the function
    """
    return {
        k: None if v.default == inspect._empty else v.default
        for k, v in inspect.signature(funct).parameters.items()
    }


def get_graph(
    obj: object,
    obj_name: Optional[str] = None,
    nodes_dict: dict = {},
    edges_lst: list = [],
    link_node: Optional[str] = None,
) -> Tuple[dict, list]:
    """
    Get dictionary of nodes with node names as keys and node objects as values. In addition, generate a list of edges,
    consisting of pairs of node names which are linked together.

    Args:
        obj (object): Object to generate dictionary of nodes and list of edges
        obj_name (str): Name of the object
        nodes_dict (dict): Dictionary of nodes
        edges_lst (list): List of edges
        link_node (str): Name of the node to link to

    Returns:
        dict, list: dictionary of nodes and list of edges
    """
    node_name = get_node_name(node=obj, node_name=obj_name)
    nodes_dict.update({node_name: obj})
    if link_node is not None:
        edges_lst.append([link_node, node_name])
    if isinstance(obj, DelayedObject):
        for k, v in obj._input.items():
            if k == "kwargs":
                for sk, sv in v.items():
                    nodes_dict, edges_lst = get_graph(
                        obj=sv,
                        obj_name=sk,
                        nodes_dict=nodes_dict,
                        edges_lst=edges_lst,
                        link_node=node_name,
                    )
            else:
                nodes_dict, edges_lst = get_graph(
                    obj=v,
                    obj_name=k,
                    nodes_dict=nodes_dict,
                    edges_lst=edges_lst,
                    link_node=node_name,
                )
    elif isinstance(obj, dict) and any(
        [isinstance(v, DelayedObject) for v in obj.values()]
    ):
        for k, v in obj.items():
            nodes_dict, edges_lst = get_graph(
                obj=v,
                obj_name=k,
                nodes_dict=nodes_dict,
                edges_lst=edges_lst,
                link_node=node_name,
            )
    elif isinstance(obj, list) and any([isinstance(v, DelayedObject) for v in obj]):
        for k, v in enumerate(obj):
            nodes_dict, edges_lst = get_graph(
                obj=v,
                obj_name=str(k),
                nodes_dict=nodes_dict,
                edges_lst=edges_lst,
                link_node=node_name,
            )
    return nodes_dict, edges_lst


def get_hash(binary: bytes) -> str:
    """
    Get the hash of a binary string - remove the specification of jupyter kernel from hash to be deterministic

    Args:
        binary (bytes): binary string to hash

    Returns:
        str: hash of the binary string
    """
    binary_no_ipykernel = re.sub(b"(?<=/ipykernel_)(.*)(?=/)", b"", binary)
    return str(hashlib.md5(binary_no_ipykernel).hexdigest())


def get_node_name(node: object, node_name: Optional[str] = None) -> str:
    """
    Get name of the node

    Args:
        node (object): Node to get the name for
        node_name (str): Name of the node in case it is already defined

    Returns:
        str: name of the node
    """
    if isinstance(node, DelayedObject) and node_name is None:
        try:
            node_name = node._function.__name__
        except TypeError:
            node_name = str(node).replace("<", "").replace(" object at ", "")
    if node_name is None:
        try:
            node_name = node.__name__
        except AttributeError:
            node_name = str(type(node))
    try:
        return node_name + "_" + get_hash(binary=cloudpickle.dumps(node))
    except TypeError:
        return node_name


def recursive_dict_resolve(input_dict: dict) -> dict:
    """
    Recursively resolve the dictionary to call result() on all objects of type DelayedObject

    Args:
        input_dict (dict): dictionary to recursively resolve

    Returns:
        dict: resolved dictionary
    """
    output_dict = {}
    for k, v in input_dict.items():
        if isinstance(v, DelayedObject):
            output_dict[k] = v.pull()
        elif isinstance(v, dict):
            output_dict[k] = recursive_dict_resolve(input_dict=v)
        elif isinstance(v, list):
            output_dict[k] = list(
                recursive_dict_resolve(
                    input_dict={k: v for k, v in enumerate(v)}
                ).values()
            )
        else:
            output_dict[k] = v
    return output_dict


class Selector:
    def __init__(self, obj: object, selector: str):
        self._obj = obj
        self._selector = selector

    def __getattr__(self, name: str):
        if self._selector == "files" and name in self._obj._output_file_lst:
            obj_copy = self._obj.__copy__()
            obj_copy._output_file = name
            return obj_copy
        elif self._selector == "output" and name in self._obj._output_key_lst:
            obj_copy = self._obj.__copy__()
            obj_copy._output_key = name
            return obj_copy
        else:
            return self.__getattribute__(name)


class DelayedObject:
    def __init__(
        self,
        function: callable,
        *args,
        output_key: Optional[str] = None,
        output_file: Optional[str] = None,
        output_file_lst: list = [],
        output_key_lst: list = [],
        list_length: Optional[int] = None,
        list_index: Optional[int] = None,
        **kwargs,
    ):
        self._input = {}
        self._function = function
        try:
            self._input.update(
                inspect.signature(self._function).bind(*args, **kwargs).arguments
            )
        except TypeError:
            pass
        self.__name__ = "DelayedObject"
        self._result = None
        self._server = Server()
        self._output_key = output_key
        self._output_file = output_file
        self._output_key_lst = output_key_lst
        self._output_file_lst = output_file_lst
        self._list_length = list_length
        self._list_index = list_index

    @property
    def server(self):
        return self._server

    def draw(self):
        node_dict, edge_lst = self.get_graph()
        draw(node_dict=node_dict, edge_lst=edge_lst)

    def get_python_result(self):
        if isinstance(self._result, dict):
            return self._result[self._output_key]
        else:
            return getattr(self._result.output, self._output_key)

    def get_file_result(self):
        return getattr(self._result.files, self._output_file)

    def pull(self):
        if self._result is None:
            self._input.update({"server_obj": self.server})
            self._result = evaluate_function(
                funct=self._function, input_dict=self._input
            )
        if self._output_key is not None:
            return self.get_python_result()
        elif self._output_file is not None:
            return self.get_file_result()
        elif self._list_index is not None:
            return self._result[self._list_index]
        else:
            return self._result

    def get_graph(self):
        return get_graph(obj=self, nodes_dict={}, edges_lst=[], link_node=None)

    def __copy__(self):
        obj_copy = DelayedObject(
            function=self._function,
            output_key=self._output_key,
            output_file=self._output_file,
            output_file_lst=self._output_file_lst,
            output_key_lst=self._output_key_lst,
            list_length=self._list_length,
            list_index=self._list_index,
        )
        obj_copy._input = self._input
        obj_copy._result = self._result
        obj_copy._server.from_dict(self._server.to_dict())
        return obj_copy

    def __getattr__(self, name):
        if name in ["files", "output"]:
            return Selector(obj=self, selector=name)
        else:
            return self.__getattribute__(name)

    def __iter__(self):
        if self._list_length is not None and self._list_index is None:
            for i in range(self._list_length):
                obj = self.__copy__()
                obj._list_index = i
                yield obj
        else:
            raise TypeError(
                "'DelayedObject' object is not iterable, when self._list_length = None."
            )



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Template class to define jobs
"""

from typing import Optional

from pyiron_base.interfaces.object import HasStorage
from pyiron_base.jobs.job.generic import GenericJob
from pyiron_base.storage.datacontainer import DataContainer

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "May 15, 2020"


class TemplateJob(GenericJob, HasStorage):
    """
    pyiron template job class for codes with an external executable.

    Example:

    >>> from pyiron_base import TemplateJob, Project

    >>> class MyJob(TemplateJob):
    >>>     def __init__(self, project, job_name):
    >>>         super().__init__(project, job_name)
    >>>         job.input.message = "hello!"
    >>>         self.executable = "cat input.dat > output.dat"

    >>>     def write_input(self):
    >>>         with open(self.working_directory + "/input.dat", "w") as f:
    >>>             f.write(job.input.message)

    >>>     def collect_output(self):
    >>>         with open(self.working_directory + "/output.dat", "w") as f:
    >>>             job.output.message = f.read()
    >>>         job.to_hdf()

    >>> pr = Project("my_project")
    >>> job = pr.create_job(MyJob, "my_job")
    >>> job.run()

    You can store information you need in `job.input` (or `self.input`) and
    `job.output` (or `self.output`). The information assigned there will be
    automatically stored in the database after a successful run. You can write
    everything inside `run_static`, but optionally you can use the functions
    `def write_input(self)` and `def collect_output(self)`, which are called
    before and after `run_static`, respectively.

    Important: The job runs in the working directory of the pyiron job. In the
    example above, it is placed under `my_project/my_job_hdf5/my_job/`. It is
    therefore important to use the absolute path, or `self.working_directory`
    to make sure that the files are found correctly

    If you have a code which requires an executable, take a look at
    :class:`~.TemplateJob` instead.

    """

    def __init__(
        self, project: "pyiron_base.storage.hdfio.ProjectHDFio", job_name: str
    ):
        GenericJob.__init__(self, project, job_name)
        HasStorage.__init__(self, group_name="")
        self.storage.create_group("input")
        self.storage.create_group("output")

    @property
    def input(self) -> DataContainer:
        return self.storage.input

    @property
    def output(self) -> DataContainer:
        return self.storage.output

    def _to_dict(self) -> dict:
        """
        Convert the job object to a dictionary.

        Returns:
            dict: A dictionary representation of the job object.
        """
        job_dict = super()._to_dict()
        job_dict["input/data"] = self.storage.input.to_dict()
        return job_dict

    def _from_dict(self, obj_dict: dict, version: str = None):
        """
        Update the object attributes from a dictionary representation.

        Args:
            obj_dict (dict): The dictionary containing the object attributes.
            version (str, optional): The version of the dictionary format. Defaults to None.
        """
        super()._from_dict(obj_dict=obj_dict, version=version)
        input_dict = obj_dict["input"]
        if "data" in input_dict.keys():
            self.storage.input.update(input_dict["data"])

    def to_hdf(
        self,
        hdf: Optional["pyiron_base.storage.hdfio.ProjectHDFio"] = None,
        group_name: Optional[str] = None,
    ) -> None:
        GenericJob.to_hdf(self=self, hdf=hdf, group_name=group_name)
        HasStorage.to_hdf(self=self, hdf=self.project_hdf5)

    def from_hdf(
        self,
        hdf: Optional["pyiron_base.storage.hdfio.ProjectHDFio"] = None,
        group_name: Optional[str] = None,
    ) -> None:
        GenericJob.from_hdf(self=self, hdf=hdf, group_name=group_name)
        HasStorage.from_hdf(self=self, hdf=self.project_hdf5)


class PythonTemplateJob(TemplateJob):
    """
    pyiron template job class for python codes.

    Example:

    >>> from pyiron_base import PythonTemplateJob, Project

    >>> class ToyJob(PythonTemplateJob):  # Create a custom job class
    >>>     def __init__(self, project, job_name):
    >>>         super().__init__(project, job_name)
    >>>         self.input.energy = 100  # Define default input

    >>>     def run_static(self):  # Call a python function and store stuff in the output
    >>>         self.output.double = self.input.energy * 2
    >>>         self.status.finished = True
    >>>         self.to_hdf()

    >>> job = pr.create_job(job_type=ToyJob, job_name="toy")  # Create job instance
    >>> job.run()  # Execute Custom job class

    You can store information you need in `job.input` (or `self.input`) and
    `job.output` (or `self.output`). The information assigned there will be
    automatically stored in the database after a successful run. You can write
    everything inside `run_static`, but optionally you can use the functions
    `def write_input(self)` and `def collect_output(self)`, whichare called
    before and after `run_static`, respectively.

    If you have a code which requires an executable, take a look at
    `TemplateJob` instead.

    """

    def __init__(
        self, project: "pyiron_base.storeage.hdfio.ProjectHDFio", job_name: str
    ):
        super().__init__(project, job_name)
        self._job_with_calculate_function = True
        self._write_work_dir_warnings = False



import os
import tarfile


def is_within_directory(directory: str, target: str) -> bool:
    """
    Check if the target path is within the specified directory.

    Args:
        directory (str): The directory path.
        target (str): The target path.

    Returns:
        bool: True if the target path is within the directory, False otherwise.
    """
    abs_directory = os.path.abspath(directory)
    abs_target = os.path.abspath(target)

    prefix = os.path.commonprefix([abs_directory, abs_target])

    return prefix == abs_directory


def safe_extract(
    tar: tarfile.TarFile,
    path: str = ".",
    members: list = None,
    *,
    numeric_owner: bool = False,
) -> None:
    """
    Safely extract the contents of a tar file.

    This function checks if the extracted files are within the specified path to prevent path traversal attacks.

    Args:
        tar (tarfile.TarFile): The tar file object.
        path (str, optional): The path to extract the files to. Defaults to ".".
        members (list, optional): The members to extract. Defaults to None.
        numeric_owner (bool, optional): Whether to use numeric owner for extracted files. Defaults to False.

    Raises:
        Exception: If attempted path traversal is detected in the tar file.
    """
    for member in tar.getmembers():
        member_path = os.path.join(path, member.name)
        if not is_within_directory(path, member_path):
            raise Exception("Attempted Path Traversal in Tar File")

    tar.extractall(path, members, numeric_owner=numeric_owner)



import os
import shutil
import tarfile
import tempfile
from typing import Optional

import pandas


def copy_files_to_archive(
    directory_to_transfer: str,
    archive_directory: str,
    compress: bool = True,
    copy_all_files: bool = False,
    arcname: Optional[str] = None,
    df: Optional[pandas.DataFrame] = None,
):
    """
    Copy files from a directory to an archive, optionally compressing the archive.

    Args:
        directory_to_transfer (str): The directory containing the files to transfer.
        archive_directory (str): The destination directory for the archive.
        compress (bool): If True, compress the archive directory into a tarball. Default is True.
        copy_all_files (bool): If True, include all files in the archive, otherwise only .h5 files. Default is False.
        arcname (str): The name of the archive directory. Default is the name of the directory to transfer.
        df (DataFrame): DataFrame containing updated job information with new IDs and project paths.

    """

    def copy_files(
        origin: str, destination: str, copy_all_files: bool = copy_all_files
    ):
        """
        Copy files from the origin directory to the destination directory.

        Args:
            origin (str): The origin directory containing the files to copy.
            destination (str): The destination directory for the copied files.
            copy_all_files (bool): If True, include all files in the archive,
                otherwise only .h5 files. Default is False.
        """
        if copy_all_files:
            shutil.copytree(origin, destination, dirs_exist_ok=True)
        else:
            copy_h5_files(origin, destination)

    assert isinstance(archive_directory, str) and ".tar.gz" not in archive_directory
    if arcname is None:
        arcname = os.path.relpath(os.path.abspath(archive_directory), os.getcwd())
    if df is not None:
        df.to_csv(os.path.join(directory_to_transfer, "export.csv"))
    if not compress:
        copy_files(directory_to_transfer, os.path.join(archive_directory, arcname))
    elif compress and copy_all_files:
        with tarfile.open(f"{archive_directory}.tar.gz", "w:gz") as tar:
            tar.add(directory_to_transfer, arcname=arcname)
    else:
        with tempfile.TemporaryDirectory() as temp_dir:
            # Copy files to the temporary directory
            dest = os.path.join(
                temp_dir, os.path.basename(directory_to_transfer.rstrip("/\\"))
            )
            copy_files(directory_to_transfer, dest)
            # Compress the temporary directory into a tar.gz archive
            with tarfile.open(f"{archive_directory}.tar.gz", "w:gz") as tar:
                tar.add(dest, arcname=arcname)
    if df is not None:
        os.remove(os.path.join(directory_to_transfer, "export.csv"))


def copy_h5_files(src: str, dst: str) -> None:
    """
    Copies all .h5 files from the source directory to the destination directory,
    preserving the directory structure.

    Args:
        src (str): The source directory from which .h5 files will be copied.
        dst (str): The destination directory where .h5 files will be copied to.

    This function traverses the source directory tree, identifies files with a .h5
    extension, and copies them to the destination directory while maintaining the
    same directory structure. Non-.h5 files are ignored.
    """

    for root, dirs, files in os.walk(src):
        for file in files:
            if file.endswith(".h5") or file == "export.csv":
                src_file = os.path.join(root, file)
                rel_path = os.path.relpath(root, src)
                dst_dir = os.path.join(dst, rel_path)
                os.makedirs(dst_dir, exist_ok=True)
                shutil.copy2(src_file, os.path.join(dst_dir, file))


def export_database(df: pandas.DataFrame) -> pandas.DataFrame:
    """
    Export the project database to an archive directory.

    Args:
        df (DataFrame): pyiron job table containing job information.

    Returns:
        DataFrame: DataFrame containing updated job information with new IDs
            and project paths.
    """

    job_translate_dict = {
        old_id: new_id for new_id, old_id in enumerate(sorted(df.id.values))
    }

    df["id"] = df["id"].map(job_translate_dict)
    df["masterid"] = df["masterid"].map(job_translate_dict)
    df["parentid"] = df["parentid"].map(job_translate_dict)
    df["project"] = df["project"].map(lambda x: os.path.relpath(x, os.getcwd()))

    df.drop(columns=["projectpath"], inplace=True)
    return df



import inspect
from typing import Optional

import cloudpickle
import numpy as np

from pyiron_base.jobs.job.template import PythonTemplateJob
from pyiron_base.project.delayed import get_function_parameter_dict, get_hash


class PythonFunctionContainerJob(PythonTemplateJob):
    """
    The PythonFunctionContainerJob is designed to wrap any kind of python function into a pyiron job object

    Example:

    >>> def test_function(a, b=8):
    >>>     return a+b
    >>>
    >>> from pyiron_base import Project
    >>> pr = Project("test")
    >>> job = pr.wrap_python_function(test_function)
    >>> job.input["a"] = 4
    >>> job.input["b"] = 5
    >>> job.run()
    >>> job.output
    >>>
    >>> test_function_wrapped = pr.wrap_python_function(test_function)
    >>> test_function_wrapped(4, b=6)
    """

    def __init__(self, project, job_name):
        super().__init__(project, job_name)
        self._function = None
        self._executor_type = None
        self._automatically_rename_on_save_using_input = False
        # Automatically rename job using function and input values at save time
        # This is useful for the edge case where these jobs are created from a wrapper
        # and automatically assigned a name based on the function name, but multiple
        # jobs are created from the same function (and thus distinguished only by their
        # input)

    @property
    def python_function(self):
        return self._function

    @python_function.setter
    def python_function(self, funct: callable) -> None:
        """
        Set the python function for the job and update the input dictionary.

        Args:
            funct (callable): The python function to be wrapped.
        """
        self.input.update(get_function_parameter_dict(funct=funct))
        self._function = funct

    def set_input(self, *args, **kwargs):
        """
        Sets the input arguments for the job.python_function function.

        Args:
            *args: Positional arguments to be passed to the function.
            **kwargs: Keyword arguments to be passed to the function.
        """
        self.input.update(
            inspect.signature(self._function).bind(*args, **kwargs).arguments
        )

    def __call__(self, *args, **kwargs):
        self.set_input(*args, **kwargs)
        self.run()
        return self.output["result"]

    def _to_dict(self) -> dict:
        """
        Convert the job object to a dictionary representation.

        Returns:
            dict: The dictionary representation of the job object.
        """
        job_dict = super()._to_dict()
        job_dict["function"] = np.void(cloudpickle.dumps(self._function))
        job_dict["_automatically_rename_on_save_using_input"] = (
            self._automatically_rename_on_save_using_input
        )
        return job_dict

    def _from_dict(self, obj_dict: dict, version: Optional[str] = None) -> None:
        """
        Load the job object from a dictionary representation.

        Args:
            obj_dict (dict): The dictionary representation of the job object.
            version (str): The version of the job object.
        """
        super()._from_dict(obj_dict=obj_dict)
        self._function = cloudpickle.loads(obj_dict["function"])
        self._automatically_rename_on_save_using_input = bool(
            obj_dict["_automatically_rename_on_save_using_input"]
        )

    def save(self) -> None:
        """
        Save the job to the project.

        If `self._automatically_rename_on_save_using_input` is True, the job name will be automatically renamed by appending
        a hash generated from the function and input arguments.

        If the job name already exists in the project, the job will be loaded from the HDF5 file and marked as finished without saving.

        Returns:
            None
        """
        if self._automatically_rename_on_save_using_input:
            self.job_name = (
                self.job_name
                + "_"
                + get_hash(
                    binary=cloudpickle.dumps(
                        {"fn": self._function, "kwargs": self.input.to_builtin()}
                    )
                )
            )

        if self.job_name in self.project.list_nodes():
            self.from_hdf()
            self.status.finished = True
            return  # Without saving
        super().save()

    def run_static(self) -> None:
        """
        Run the static function.

        If an executor is specified and the function signature contains an 'executor' parameter,
        the function is executed using the specified executor. Otherwise, the function is executed
        without an executor.
        """
        self.status.running = True
        if (
            self._executor_type is not None
            and "executor" in inspect.signature(self._function).parameters.keys()
        ):
            input_dict = self.input.to_builtin()
            del input_dict["executor"]
            with self._get_executor(max_workers=self.server.cores) as exe:
                output = self._function(**input_dict, executor=exe)
        else:
            output = self._function(**self.input)
        self.output.update({"result": output})
        self.to_hdf()
        self.status.finished = True



from typing import Optional

import cloudpickle
import numpy as np

from pyiron_base.jobs.job.runfunction import (
    CalculateFunctionCaller,
    write_input_files_from_input_dict,
)
from pyiron_base.jobs.job.template import TemplateJob


class ExecutableContainerJob(TemplateJob):
    """
    The ExecutableContainerJob is designed to wrap any kind of external executable into a pyiron job object by providing
    a write_input(input_dict, working_directory) and a collect_output(working_directory) function.

    Example:

    >>> import os
    >>>
    >>> def write_input(input_dict, working_directory="."):
    >>>     with open(os.path.join(working_directory, "input_file"), "w") as f:
    >>>         f.write(str(input_dict["energy"]))
    >>>
    >>>
    >>> def collect_output(working_directory="."):
    >>>     with open(os.path.join(working_directory, "output_file"), "r") as f:
    >>>         return {"energy": float(f.readline())}
    >>>
    >>>
    >>> from pyiron_base import Project
    >>> pr = Project("test")
    >>> pr.create_job_class(
    >>>     class_name="CatJob",
    >>>     write_input_funct=write_input,
    >>>     collect_output_funct=collect_output,
    >>>     default_input_dict={"energy": 1.0},
    >>>     executable_str="cat input_file > output_file",
    >>> )
    >>> job = pr.create.job.CatJob(job_name="job_test")
    >>> job.input["energy"] = 2.0
    >>> job.run()
    >>> print(job.output)
    DataContainer({'energy': 2.0})

    """

    def __init__(self, project, job_name):
        super().__init__(project, job_name)
        self._write_input_funct = None
        # Set job_with_calculate_function flag to true to use run_static() to execute the python function generated by
        # the job with its arguments job.get_calculate_function(**job.calculate_kwargs) without calling the old
        # interface with write_input() and collect_output(). Finally, the output dictionary is stored in the HDF5 file
        # using self.save_output(output_dict, shell_output)
        self._job_with_calculate_function = True

    @property
    def calculate_kwargs(self) -> dict:
        """
        Generate keyword arguments for the calculate() function.

        Example:

        >>> calculate_function = job.get_calculate_function()
        >>> shell_output, parsed_output, job_crashed = calculate_function(**job.calculate_kwargs)
        >>> job.save_output(output_dict=parsed_output, shell_output=shell_output)

        Returns:
            dict: keyword arguments for the calculate() function
        """
        kwargs = super().calculate_kwargs
        kwargs.update(
            {
                "input_parameter_dict": self.input.to_builtin(),
                "executable_script": self.executable.executable_path,
                "shell_parameter": True,
            }
        )
        return kwargs

    def set_job_type(
        self,
        executable_str: str,
        write_input_funct: callable = None,
        collect_output_funct: callable = None,
        default_input_dict: dict = None,
    ):
        """
        Set the pre-defined write_input() and collect_output() function plus a dictionary of default inputs and an
        executable string.

        Args:
            executable_str (str): Call to an external executable
            write_input_funct (callable): The write input function write_input(input_dict, working_directory)
            collect_output_funct (callable): The collect output function collect_output(working_directory)
            default_input_dict (dict/None): Default input for the newly created job class

        Returns:
            callable: Function which requires a project and a job_name as input and returns a job object
        """
        self.executable = executable_str
        if write_input_funct is not None:
            self._write_input_funct = write_input_funct
        if collect_output_funct is not None:
            self._collect_output_funct = collect_output_funct
        if default_input_dict is not None:
            self.input.update(default_input_dict)

    def get_calculate_function(self) -> callable:
        """
        Generate calculate() function

        Example:

        >>> calculate_function = job.get_calculate_function()
        >>> shell_output, parsed_output, job_crashed = calculate_function(**job.calculate_kwargs)
        >>> job.save_output(output_dict=parsed_output, shell_output=shell_output)

        Returns:
            callable: calculate() functione
        """

        def get_combined_write_input_funct(
            input_job_dict: dict, write_input_funct: Optional[callable] = None
        ):
            def write_input_combo_funct(working_directory: str, input_dict: dict):
                write_input_files_from_input_dict(
                    input_dict=input_job_dict,
                    working_directory=working_directory,
                )
                if write_input_funct is not None:
                    write_input_funct(
                        working_directory=working_directory,
                        input_dict=input_dict,
                    )

            return write_input_combo_funct

        return CalculateFunctionCaller(
            write_input_funct=get_combined_write_input_funct(
                input_job_dict=self.get_input_parameter_dict(),
                write_input_funct=self._write_input_funct,
            ),
            collect_output_funct=self._collect_output_funct,
        )

    def _to_dict(self) -> dict:
        """
        Convert the job object to a dictionary representation.

        Returns:
            dict: A dictionary representation of the job object.
        """
        job_dict = super()._to_dict()
        if self._write_input_funct is not None:
            job_dict["write_input_function"] = np.void(
                cloudpickle.dumps(self._write_input_funct)
            )
        if self._collect_output_funct is not None:
            job_dict["collect_output_function"] = np.void(
                cloudpickle.dumps(self._collect_output_funct)
            )
        return job_dict

    def _from_dict(self, obj_dict: dict, version: str = None):
        """
        Load the job attributes from a dictionary representation.

        Args:
            obj_dict (dict): A dictionary containing the job attributes.
            version (str): The version of the job object.

        """
        super()._from_dict(obj_dict=obj_dict)
        if "write_input_function" in obj_dict.keys():
            self._write_input_funct = cloudpickle.loads(
                obj_dict["write_input_function"]
            )
        if "write_input_function" in obj_dict.keys():
            self._collect_output_funct = cloudpickle.loads(
                obj_dict["collect_output_function"]
            )



from typing import Callable, Dict, Optional

from pyiron_base.utils.instance import static_isinstance


def create_job_factory(
    executable_str: str,
    write_input_funct: Optional[Callable] = None,
    collect_output_funct: Optional[Callable] = None,
    default_input_dict: Optional[Dict] = None,
) -> callable:
    """
    Create a new job class based on pre-defined write_input() and collect_output() function plus a dictionary of
    default inputs and an executable string.

    Args:
        executable_str (str): Call to an external executable
        write_input_funct (callable): The write input function write_input(input_dict, working_directory)
        collect_output_funct (callable): The collect output function collect_output(working_directory)
        default_input_dict (dict/None): Default input for the newly created job class

    Example:

    >>> def write_input(input_dict, working_directory="."):
    >>>     with open(os.path.join(working_directory, "input_file"), "w") as f:
    >>>         f.write(str(input_dict["energy"]))
    >>>
    >>>
    >>> def collect_output(working_directory="."):
    >>>     with open(os.path.join(working_directory, "output_file"), "r") as f:
    >>>         return {"energy": float(f.readline())}
    >>>
    >>>
    >>> from pyiron_base import Project, create_job_factory
    >>> pr = Project("test")
    >>> create_catjob = create_job_factory(
    >>>     write_input_funct=write_input,
    >>>     collect_output_funct=collect_output,
    >>>     default_input_dict={"energy": 1.0},
    >>>     executable_str="cat input_file > output_file",
    >>> )
    >>> job = create_catjob(project=pr, job_name="job_test")
    >>> job.input["energy"] = 2.0
    >>> job.run()
    >>> job.output
    """

    def job_factory(project, job_name):
        """
        Create a job based on the previously defined write_input(), collect_output() and the executable string.

        Args:
            project (ProjectHDFio/ Project): ProjectHDFio instance which points to the HDF5 file the job is stored in
            job_name (str): name of the job, which has to be unique within the project

        Returns:
            pyiron_base.jobs.flex.executablecontainer.ExecutableContainerJob: pyiron job object
        """
        if static_isinstance(project, "pyiron_base.project.generic.Project"):
            job = project.create.job.ExecutableContainerJob(job_name=job_name)
        elif static_isinstance(project, "pyiron_base.storage.hdfio.ProjectHDFio"):
            job = project.project.create.job.ExecutableContainerJob(job_name=job_name)
        else:
            raise TypeError(
                "Expected ProjectHDFio/ Project but recieved", type(project)
            )
        job.set_job_type(
            write_input_funct=write_input_funct,
            collect_output_funct=collect_output_funct,
            default_input_dict=default_input_dict,
            executable_str=executable_str,
        )
        return job

    return job_factory



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
DatabaseAccess class deals with accessing the database
"""

import fnmatch
import re
import warnings
from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import List, Literal, Optional, Union

import numpy as np
import pandas
from pyiron_snippets.deprecate import deprecate
from pyiron_snippets.logger import logger

__author__ = "Murat Han Celik"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH"
    " - Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


class IsDatabase(ABC):
    """
    Captures common interface for all database types in pyiron, e.g. SQL/SQLite/FileTable.
    """

    @abstractmethod
    def _get_view_mode(self) -> bool:
        pass

    @property
    def view_mode(self) -> bool:
        """
        Get view_mode - if view_moded is enable pyiron has read only access to the database.

        Some implementations do not allow to set this value.

        Returns:
            bool: True when view_mode is enabled
        """
        return self._get_view_mode()

    @property
    @deprecate("use view_mode")
    def viewer_mode(self) -> bool:
        return self.view_mode

    viewer_mode.__doc__ = view_mode.__doc__

    @abstractmethod
    def _get_job_table(
        self,
        sql_query: str,
        user: str,
        project_path: str,
        recursive: bool = True,
        columns: Optional[List[str]] = None,
        element_lst: Optional[List[str]] = None,
    ) -> pandas.DataFrame:
        pass

    @staticmethod
    def _get_filtered_job_table(
        df: pandas.DataFrame,
        mode: Literal["regex", "glob"] = "glob",
        **kwargs: dict,
    ) -> pandas.DataFrame:
        """
        Get a job table in a project based on matching values from any column in the project database

        The values in `kwargs` can be wildcards. The matches can be given
        either via "glob" or "regex".

        Args:
            df (pandas.DataFrame): DataFrame to be filtered
            **kwargs (dict): Optional arguments for filtering with keys matching the project database column name
                            (eg. status="finished")

        Returns:
            list: DataFrame containing filtered jobs
        """
        if len(kwargs) == 0 or df.empty:
            return df
        mask = np.ones_like(df.index, dtype=bool)
        for key in kwargs.keys():
            if key not in list(df.columns):
                raise ValueError(
                    f"Column name {key} does not exist in the project database!"
                )
        for key, val in kwargs.items():
            if mode == "regex":
                pattern = re.compile(str(val))
                update = df[key].apply(pattern.search).astype(bool)
            elif mode == "glob":
                if str(val).startswith("!"):
                    logger.warn(
                        "It looks like you are using an old pyiron convention."
                        " If you meant to exclude the term following '!', use"
                        " `mode='regex' and use a regex convention (such as"
                        " `^(?!term$)`)"
                    )
                arr = np.asarray(df[key]).astype(str)
                matches = fnmatch.filter(arr, str(val))
                update = np.array([k in matches for k in arr])
            mask &= update
        return df[mask]

    def job_table(
        self,
        sql_query: str,
        user: str,
        project_path: str,
        recursive: bool = True,
        columns: Optional[List[str]] = None,
        all_columns: bool = False,
        sort_by: str = "id",
        max_colwidth: int = 200,
        full_table: bool = False,
        element_lst: Optional[List[str]] = None,
        job_name_contains: str = "",
        mode: Literal["regex", "glob"] = "glob",
        **kwargs,
    ):
        """
        Access the job_table.

        Args:
            sql_query (str): SQL query to enter a more specific request
            user (str): username of the user whoes user space should be searched
            project_path (str): root_path - this is in contrast to the project_path in GenericPath
            recursive (bool): search subprojects [True/False]
            columns (list): by default only the columns ['job', 'project', 'chemicalformula'] are selected, but the
                            user can select a subset of ['id', 'status', 'chemicalformula', 'job', 'subjob', 'project',
                            'projectpath', 'timestart', 'timestop', 'totalcputime', 'computer', 'hamilton', 'hamversion',
                            'parentid', 'masterid']
            all_columns (bool): Select all columns - this overwrites the columns option.
            sort_by (str): Sort by a specific column
            max_colwidth (int): set the column width
            full_table (bool): Whether to show the entire pandas table
            element_lst (list): list of elements required in the chemical formular - by default None
            job_name_contains (str): (deprecated) A string which should be contained in every job_name
            mode (str): search mode when kwargs are given.
            **kwargs (dict): Optional arguments for filtering with keys matching the project database column name
                            (eg. status="finished"). Asterisk can be used to denote a wildcard, for zero or more
                            instances of any character

        Returns:
            pandas.Dataframe: Return the result as a pandas.Dataframe object
        """
        if columns is None:
            columns = ["job", "project", "chemicalformula"]
        if all_columns:
            columns = [
                "id",
                "status",
                "chemicalformula",
                "job",
                "subjob",
                "projectpath",
                "project",
                "timestart",
                "timestop",
                "totalcputime",
                "computer",
                "hamilton",
                "hamversion",
                "parentid",
                "masterid",
            ]
        if sort_by not in columns:
            columns = list(columns) + [sort_by]
        if full_table:
            pandas.set_option("display.max_rows", None)
            pandas.set_option("display.max_columns", None)
        else:
            pandas.reset_option("display.max_rows")
            pandas.reset_option("display.max_columns")
        pandas.set_option("display.max_colwidth", max_colwidth)
        df = self._get_job_table(
            user=user,
            sql_query=sql_query,
            project_path=project_path,
            recursive=recursive,
            columns=columns,
        )
        if job_name_contains != "":
            warnings.warn(
                "`job_name_contains` is deprecated - use `job='*term*'` instead"
            )
            kwargs["job"] = "*{}*".format(job_name_contains)
        df = self._get_filtered_job_table(df, mode=mode, **kwargs)
        if sort_by is not None:
            return df.sort_values(by=sort_by)
        return df

    @abstractmethod
    def _get_table_headings(self, table_name: Optional[str] = None):
        pass

    def item_update(self, par_dict: dict, item_id: int) -> None:
        if isinstance(item_id, Iterable):
            return self._items_update(par_dict=par_dict, item_ids=item_id)
        return self._item_update(par_dict=par_dict, item_id=item_id)

    @abstractmethod
    def _item_update(self, par_dict: dict, item_id: int) -> None:
        pass

    def _items_update(self, par_dict: dict, item_ids: List) -> None:
        """
        For now simply loops over all item_ids to call item_update,
        but can be made more efficient.
        Should be made an asbtract method when defined in inheriting classes

        Args:
            par_dict (_type_): _description_
            item_ids (_type_): _description_
        """
        for i_id in item_ids:
            self._item_update(par_dict=par_dict, item_id=i_id)

    def set_job_status(self, status: str, job_id: Union[int, List[int]]) -> None:
        """
        Set status of a job or multiple jobs if job_id is iterable.

        Args:
            status (str): status
            job_id (int, Iterable): job id
        """
        if isinstance(job_id, Iterable):
            return self._items_update(
                par_dict={"status": status},
                item_ids=job_id,
            )
        return self._item_update(
            par_dict={"status": status},
            item_id=job_id,
        )

    def get_table_headings(self, table_name: Optional[str] = None) -> List[str]:
        """
        Get column names; if given table_name can select one of multiple tables defined in the database, but subclasses
        may ignore it

        Args:
            table_name (str): simple string of a table_name like: 'jobs_username'

        Returns:
            list: list of column names like:
                ['id',
                'parentid',
                'masterid',
                'projectpath',
                'project',
                'job',
                'subjob',
                'chemicalformula',
                'status',
                'hamilton',
                'hamversion',
                'username',
                'computer',
                'timestart',
                'timestop',
                'totalcputime']
        """
        return self._get_table_headings(table_name=table_name)

    @deprecate("use get_table_headings()")
    def get_db_columns(self) -> List[str]:
        """
        Get column names

        Returns:
            list: list of column names like:
                ['id',
                'parentid',
                'masterid',
                'projectpath',
                'project',
                'job',
                'subjob',
                'chemicalformula',
                'status',
                'hamilton',
                'hamversion',
                'username',
                'computer',
                'timestart',
                'timestop',
                'totalcputime']
        """
        return self.get_table_headings()

    @abstractmethod
    def _get_jobs(
        self,
        sql_query: str,
        user: str,
        project_path: str,
        recursive: bool = True,
        columns: Optional[List[str]] = None,
    ) -> List[dict]:
        pass

    def get_jobs(
        self,
        sql_query: str,
        user: str,
        project_path: str,
        recursive: bool = True,
        columns: Optional[List[str]] = None,
    ) -> List[dict]:
        """
        Internal function to return the jobs as dictionary rather than a pandas.Dataframe

        Args:
            sql_query (str): SQL query to enter a more specific request
            user (str): username of the user whoes user space should be searched
            project_path (str): root_path - this is in contrast to the project_path in GenericPath
            recursive (bool): search subprojects [True/False]
            columns (list): by default only the columns ['id', 'project'] are selected, but the user can select a subset
                            of ['id', 'status', 'chemicalformula', 'job', 'subjob', 'project', 'projectpath', 'timestart',
                            'timestop', 'totalcputime', 'computer', 'hamilton', 'hamversion', 'parentid', 'masterid']

        Returns:
            dict: columns are used as keys and point to a list of the corresponding values
        """
        if columns is None:
            columns = ["id", "project"]
        return self._get_jobs(sql_query, user, project_path, recursive, columns)

    def get_job_ids(
        self, sql_query: str, user: str, project_path: str, recursive: bool = True
    ) -> List[int]:
        """
        Return the job IDs matching a specific query

        Args:
            database (DatabaseAccess): Database object
            sql_query (str): SQL query to enter a more specific request
            user (str): username of the user whoes user space should be searched
            project_path (str): root_path - this is in contrast to the project_path in GenericPath
            recursive (bool): search subprojects [True/False]

        Returns:
            list: a list of job IDs
        """
        return self.get_jobs(
            sql_query=sql_query,
            user=user,
            project_path=project_path,
            recursive=recursive,
        )["id"]



"""
Backwards compatible way of importing the DataContainer.
"""

from typing import Optional, Union

from pyiron_snippets.deprecate import deprecate

from pyiron_base.storage.datacontainer import DataContainer


class InputList(DataContainer):
    """
    A class representing an input list.

    This class is deprecated. Please use DataContainer instead.

    Args:
        init (Union[None, dict, list, tuple], optional): The initial data to populate the input list with. Defaults to None.
        table_name (str, optional): The name of the table associated with the input list. Defaults to None.
    """

    @deprecate("use DataContainer instead", version="0.3.0")
    def __init__(
        self,
        init: Union[None, dict, list, tuple] = None,
        table_name: Optional[str] = None,
    ) -> None:
        """
        Initialize the InputList object.

        Args:
            init (Union[None, dict, list, tuple], optional): The initial data to populate the input list with. Defaults to None.
            table_name (str, optional): The name of the table associated with the input list. Defaults to None.
        """
        super().__init__(init=init, table_name=table_name)



import math
import os

import pint


def _size_conversion(size: pint.Quantity):
    sign_prefactor = 1
    if size < 0:
        sign_prefactor = -1
        size *= -1
    elif size == 0:
        return size

    prefix_index = math.floor(math.log2(size) / 10) - 1
    prefix = ["Ki", "Mi", "Gi", "Ti", "Pi"]

    size *= sign_prefactor
    if prefix_index < 0:
        return size
    elif prefix_index < 5:
        return size.to(f"{prefix[prefix_index]}byte")
    else:
        return size.to(f"{prefix[-1]}byte")


def get_folder_size(path):
    size = (
        sum(
            [
                sum([os.path.getsize(os.path.join(path, f)) for f in files])
                for path, dirs, files in os.walk(path)
            ]
        )
        * pint.UnitRegistry().byte
    )
    return _size_conversion(size)



# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Remove jobs from pyiron project or whole project.
"""

import os
from argparse import ArgumentParser, Namespace

from pyiron_base.project.generic import Project

__author__ = "Marvin Poul"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Marvin Poul"
__email__ = "poul@mpie.de"
__status__ = "production"
__date__ = "23 Jun, 2020"


def register(parser: ArgumentParser):
    parser.add_argument(
        "project", default=".", nargs="?", help="path to pyiron project"
    )
    parser.add_argument(
        "-j", "--jobs-only", action="store_true", help="only remove jobs inside project"
    )
    parser.add_argument(
        "-r", "--recursive", action="store_true", help="recurse into subprojects"
    )


def main(args: Namespace) -> None:
    pr = Project(args.project)
    if args.jobs_only:
        pr.remove_jobs(recursive=args.recursive, silently=True)
    else:
        pr.remove(enable=True)
        if os.path.exists(args.project) and not os.listdir(args.project):
            os.rmdir(args.project)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Load input parameters for jupyter notebooks from external HDF5 or JSON file
"""

import json
import warnings
from pathlib import Path

from pyiron_base.storage.datacontainer import DataContainer
from pyiron_base.storage.hdfio import FileHDFio

__author__ = "Osamu Waseda"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2019"


class Notebook(object):
    """
    class for pyiron notebook objects

    # TODO: Extract JSON functionality over to Project.data.read/write and remove this file.
    """

    @staticmethod
    def get_custom_dict():
        return load()

    @staticmethod
    def store_custom_output_dict(output_dict):
        dump(output_dict=output_dict)


def load() -> DataContainer:
    """
    Load input parameters from HDF5 or JSON file.

    Returns:
        DataContainer: The loaded input parameters.
    """
    folder = Path(".").cwd().parts[-1]
    project_folder = Path(".").cwd().parents[1]
    hdf_file = project_folder / folder
    hdf_file = str(hdf_file).replace("\\", "/") + ".h5"
    if Path(hdf_file).exists():
        obj = DataContainer()
        hdf_file_obj = FileHDFio(hdf_file)
        hdf_input = hdf_file_obj[folder + "/input"]
        if "custom_dict" in hdf_input.list_nodes():
            obj.update(hdf_file_obj[folder + "/input/custom_dict"])
        else:  # Backwards compatibility
            obj.from_hdf(hdf=hdf_file_obj, group_name=folder + "/input/custom_dict")
        obj["project_dir"] = str(project_folder)
        return obj
    elif Path("input.json").exists():
        with open("input.json") as f:
            return json.load(f)
    else:
        warnings.warn("{} not found".format(hdf_file))
        return None


def dump(output_dict: dict) -> None:
    """
    Dump output dictionary to HDF5 file.

    Args:
        output_dict (dict): The output dictionary to be dumped.
    """
    folder = Path(".").cwd().parts[-1]
    hdf_file = Path(".").cwd().parents[1] / folder
    hdf_file = str(hdf_file) + ".h5"
    hdf = FileHDFio(hdf_file)
    hdf[folder].create_group("output")
    obj = DataContainer(output_dict)
    obj.to_hdf(hdf[folder + "/output"])



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Copy a pyiron project to a new location.
"""

from argparse import ArgumentParser, Namespace

from pyiron_base.project.generic import Project


def register(parser: ArgumentParser) -> None:
    parser.add_argument("src", help="source project")
    parser.add_argument("dst", help="destination project")


def main(args: Namespace) -> None:
    src = Project(args.src)
    dst = Project(args.dst)
    src.copy_to(dst)
